""" 
Copyright <2020> <LENA SIMINE>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 
If you have any questions of comments please contact: lena.simine at mcgill dot ca 
If you use this code please cite this paper:


This code was used to train a Long-Short-Term-Memory Recurrent Neural Network 
that relate the torsional profiles of Poly-3-hexylthiophene 30-mers 
to their absorption spectra back when python 2.7 and tensorflow <2.0 were popular.
The code accepts two text files as input: 'input' and 'output', does pre-processing,
and produces a model that is saved in checkpoint files. The paths to files 
need to be adjusted to your local environment throughout.


"""

from __future__ import print_function
import time
import sys
import tensorflow as tf
from tensorflow.contrib import rnn
import numpy as np
from numpy import genfromtxt

features = genfromtxt('input')
labels = genfromtxt('output')

# Training Parameters
learning_rate = 0.05 
training_steps = 30000
batch_size = 300 
display_step = 1000

# Network Parameters
Nchrom=6
num_input = Nchrom 
timesteps = 29-Nchrom  
num_hidden = 150 
num_classes = 20 #this needs to match num_classes in the definition of  dense_to_one_hot function below

#dataset pre-processing
#turn the lists of dihedrals into the input sequence using a moving window (window length: Nchrom) 
features_lstm=np.zeros(shape=(np.shape(features)[0],Nchrom*(29-Nchrom)), dtype=np.int32) 

for ii in range(np.shape(features)[0]):
  temp=[]
  for jj in range(29-Nchrom):
    temp.append(features[ii,jj:jj+Nchrom])
  features_lstm[ii,:]=np.reshape(np.array(temp),Nchrom*(29-Nchrom))

features_lstm=features_lstm


def dense_to_one_hot(labels_dense, num_classes=20):
  #Convert class labels from scalars to one-hot vectors."""
  num_labels = labels_dense.shape[0]
  index_offset = np.arange(num_labels) * num_classes
  labels_one_hot = np.zeros((num_labels, num_classes), dtype=np.int32)
  for ii in range(num_labels):
    for jj in range(num_classes):
      if jj==labels_dense[ii]:
        labels_one_hot[ii,jj]=np.int32(1)
  return labels_one_hot
def extract_labels(one_hot):
  if one_hot:
    return dense_to_one_hot(labels)
  return labels

#convert my labels to one_hot format (1 in the position defined by my label, 0's everywhere else)
oh_labels=extract_labels(one_hot=True)

indexes=np.arange(labels.shape[0])
index_test=indexes[1::10].tolist()
index_tr = [i for i in range(indexes.size) if i not in index_test]

x_train = features_lstm[index_tr,:]
y_train = oh_labels[index_tr,:]
x_test = features_lstm[index_test,:]
x_test_out=features[index_test,:]

y_test = oh_labels[index_test,:]


def iterate_minibatches(inputs, targets, batchsize, shuffle):
    assert len(inputs) == len(targets)
    if shuffle:
        indices = np.arange(len(inputs))
        np.random.shuffle(indices)
 
    for start_idx in range(0, len(inputs) - batchsize + 1, batchsize):
        if shuffle:
            excerpt = indices[start_idx:start_idx + batchsize]
        else:
            excerpt = slice(start_idx, start_idx + batchsize)
        yield inputs[excerpt], targets[excerpt]


# tf Graph input
X = tf.placeholder("float", [None, timesteps, num_input])
Y = tf.placeholder("float", [None, num_classes])

# Define weights
weights = {
    'out': tf.Variable(tf.random_normal([num_hidden, num_classes]))
}
biases = {
    'out': tf.Variable(tf.random_normal([num_classes]))
}


def RNN(x, weights, biases):

    # Prepare data shape to match `rnn` function requirements
    # Current data input shape: (batch_size, timesteps, n_input)
    # Required shape: 'timesteps' tensors list of shape (batch_size, n_input)
    # Unstack to get a list of 'timesteps' tensors of shape (batch_size, num_input)
    x = tf.unstack(x, timesteps, 1)

    # Define a lstm cell with tensorflow
    lstm_cell = rnn.BasicLSTMCell(num_hidden, forget_bias=1.0)

    # Get lstm cell output
    outputs, states = rnn.static_rnn(lstm_cell, x, dtype=tf.float32)

    # Linear activation, using rnn inner loop last output
    return tf.matmul(outputs[-1], weights['out']) + biases['out']

logits = RNN(X, weights, biases)
prediction = tf.nn.softmax(logits)

# Define loss and optimizer
loss_op = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(
    logits=logits, labels=Y))
optimizer = tf.train.GradientDescentOptimizer(learning_rate=learning_rate)
train_op = optimizer.minimize(loss_op)

# Evaluate model (with test logits, for dropout to be disabled)
correct_pred = tf.equal(tf.argmax(prediction, 1), tf.argmax(Y, 1))
accuracy = tf.reduce_mean(tf.cast(correct_pred, tf.float32))

# Initialize the variables (i.e. assign their default value)
init = tf.global_variables_initializer()

# Start training
with tf.Session() as sess:

    # Run the initializer
    sess.run(init)
    for step in range(1, training_steps+1):
        if step % display_step == 1 and step>500:
          t0=time.time()
        total_batch = int(labels.shape[0]/batch_size)
        # Loop over all batches
        count=0
        for i in iterate_minibatches(x_train, y_train, batch_size, shuffle=True):
            count=count+1
            batch_x, batch_y = i
            if count>0:
                break
        # Reshape data 
        batch_x = batch_x.reshape((batch_size, timesteps, num_input))
        # Run optimization op (backprop)
        sess.run(train_op, feed_dict={X: batch_x, Y: batch_y})
        if step % display_step == 0 or step == 1:
            # Calculate batch loss and accuracy
            loss, acc = sess.run([loss_op, accuracy], feed_dict={X: batch_x,
                                                                 Y: batch_y})
            print("Step " + str(step) + ", Minibatch Loss= " + \
                  "{:.4f}".format(loss) + ", Training Accuracy= " + \
                  "{:.3f}".format(acc))
            if step > 1500:
              t1=time.time()
              print("time of 1000 iterations = ", t1-t0)
    saver = tf.train.Saver()
    # Calculate accuracy
    test_len = np.shape(y_test)[0]

    #adjust the shape of test inputs
    x_test = x_test.reshape(test_len, timesteps, num_input)
    print("Testing Accuracy:", \
        sess.run(accuracy, feed_dict={X: x_test, Y: y_test}))

    results=sess.run(prediction, feed_dict={X: x_test, Y: y_test})
    np.savetxt('model_test_results', results)
    np.savetxt('true_test_results',y_test )
    save_path=saver.save(sess,'~/model/model.ckpt')
    
